/*
   rdesktop: A Remote Desktop Protocol client.
   Cache routines
   Copyright (C) Matthew Chapman 1999-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <assert.h>
#include "rdesktop.h"
#include "gui.h"
#include "cache.h"

#define NUM_ELEMENTS(array) (sizeof(array) / sizeof(array[0]))


/* BITMAP CACHE */
static HBITMAP bmpcache[BMP_CACHECNT][BMP_CACHESIZE];

/* Create bitmap cache */
void cache_bmp_init()
{
	int cache, idx; /* Loop counters */

	for(cache=0; cache < BMP_CACHECNT; cache++)
		for(idx=0; idx < BMP_CACHESIZE; idx++)
			bmpcache[cache][idx] = NULL;
}

/* Destroy bitmap cache */
void cache_bmp_destroy()
{
	int cache, idx; /* Loop counters */
	int delcnt = 0;

	for(cache=0; cache < BMP_CACHECNT; cache++)
		for(idx=0; idx < BMP_CACHESIZE; idx++)
			if (NULL != bmpcache[cache][idx])
			{
				ui_destroy_bitmap(bmpcache[cache][idx]);
				bmpcache[cache][idx] = NULL;
				delcnt++;
			}

	#ifdef CACHE_DEBUG
	printf("cache_bmp_destroy(): free %d elements\n", delcnt);
	#endif

	cache_bmp_init();
}


/* Retrieve a bitmap from the cache */
HBITMAP cache_bmp_get(uint8 cache_id, uint16 cache_idx)
{
	HBITMAP bitmap;

	if ((cache_id < NUM_ELEMENTS(bmpcache))
			&& (cache_idx < NUM_ELEMENTS(bmpcache[0])))
	{
		bitmap = bmpcache[cache_id][cache_idx];
		if (bitmap != NULL)
			return bitmap;
	}

	ERROR("get bitmap %d:%d\n", cache_id, cache_idx);
	return NULL;
}

/* Store a bitmap in the cache */
void cache_bmp_put(uint8 cache_id, uint16 cache_idx, HBITMAP bitmap)
{
	HBITMAP old;

	if ((cache_id < BMP_CACHECNT)
			&& (cache_idx < BMP_CACHESIZE))
	{
		old = bmpcache[cache_id][cache_idx];
		if (old != NULL)
		{
			ui_destroy_bitmap(old);
			bmpcache[cache_id][cache_idx] = NULL;
		}

		bmpcache[cache_id][cache_idx] = bitmap;
		gdk_pixmap_ref(bitmap);
	}
	else
	{
		ERROR("put bitmap %d:%d\n", cache_id, cache_idx);
		assert(BMP_CACHE_ERROR_PUT_TO_NONEXISTANT_LOCATION);
	}
}




/* FONT CACHE */
static FONTGLYPH fontcache[FONT_CACHECNT][FONT_CACHESIZE];

/* Create font cache */
void cache_font_init()
{
	int cache, idx; /* Loop counters */

	for(cache=0; cache < FONT_CACHECNT; cache++)
		for(idx=0; idx < FONT_CACHESIZE; idx++)
			fontcache[cache][idx].pixmap = NULL;
}

/* Destroy font cache */
void cache_font_destroy()
{
	int cache, idx; /* Loop counters */
	int delcnt = 0;

	for(cache=0; cache < FONT_CACHECNT; cache++)
		for(idx=0; idx < FONT_CACHESIZE; idx++)
			if(NULL != fontcache[cache][idx].pixmap)
			{
				ui_destroy_glyph(fontcache[cache][idx].pixmap);
				delcnt++;
			}

	#ifdef CACHE_DEBUG
	printf("cache_font_destroy(): free %d elements\n", delcnt);
	#endif
	cache_font_init();
}


/* Retrieve a glyph from the font cache */
FONTGLYPH *cache_font_get(uint8 font, uint16 character)
{
	FONTGLYPH *glyph;

	if ((font < NUM_ELEMENTS(fontcache))
			&& (character < NUM_ELEMENTS(fontcache[0])))
	{
		glyph = &fontcache[font][character];
		if (glyph->pixmap != NULL)
			return glyph;
	}

	ERROR("get font %d:%d\n", font, character);
	return NULL;
}

/* Store a glyph in the font cache */
void cache_font_put(uint8 font, uint16 character, uint16 offset,
		uint16 baseline, uint16 width, uint16 height, HGLYPH pixmap)
{
	FONTGLYPH *glyph;

	if ((font < NUM_ELEMENTS(fontcache))
			&& (character < NUM_ELEMENTS(fontcache[0])))
	{
		glyph = &fontcache[font][character];
		if (glyph->pixmap != NULL)
			ui_destroy_glyph(glyph->pixmap);

		glyph->offset = offset;
		glyph->baseline = baseline;
		glyph->width = width;
		glyph->height = height;
		glyph->pixmap = pixmap;
		gdk_pixmap_ref(glyph->pixmap);
	}
	else
	{
		ERROR("put font %d:%d\n", font, character);
		assert(FONT_CACHE_ERROR_PUT_TO_NONEXISTANT_LOCATION);
	}
}


/* TEXT CACHE */
static DATABLOB textcache[256];

/* Create text cache */
void cache_text_init()
{
	int idx; /* Loop counters */

	for(idx=0; idx < TEXT_CACHESIZE; idx++)
	{
		textcache[idx].data = NULL;
		textcache[idx].size = 0;
	}
}

/* Destroy font cache */
void cache_text_destroy()
{
	int idx; /* Loop counters */
	int delcnt = 0;

	for(idx=0; idx < TEXT_CACHESIZE; idx++)
	{
		if (NULL != textcache[idx].data)
		{
			g_free(textcache[idx].data);
			delcnt++;
		}
	}

	#ifdef CACHE_DEBUG
	printf("cache_text_destroy(): free %d elements\n", delcnt);
	#endif

	cache_text_init();
}


/* Retrieve a text item from the cache */
DATABLOB *cache_text_get(uint8 cache_id)
{
	DATABLOB *text;

	if (cache_id < NUM_ELEMENTS(textcache))
	{
		text = &textcache[cache_id];
		if (text->data != NULL)
			return text;
	}

	ERROR("get text %d\n", cache_id);
	return NULL;
}

/* Store a text item in the cache */
void cache_text_put(uint8 cache_id, void *data, int length)
{
	DATABLOB *text;

	if (cache_id < NUM_ELEMENTS(textcache))
	{
		text = &textcache[cache_id];
		/*if (text->data != NULL)
			g_free(text->data);*/

		text->data = g_malloc(length); assert(NULL != text->data);
		text->size = length;
		memcpy(text->data, data, length);
	}
	else
	{
		ERROR("put text %d\n", cache_id);
		assert(TEXT_CACHE_ERROR_PUT_TO_NONEXISTANT_LOCATION);
	}
}


/* DESKTOP CACHE */
static GdkPixmap* deskcache[DESK_CACHESIZE];

/* Create text cache */
void cache_desktop_init()
{
	int idx; /* Loop counters */

	for(idx=0; idx < DESK_CACHESIZE; idx++)
			deskcache[idx] = NULL;
}

/* Destroy font cache */
void cache_desktop_destroy()
{
	int idx; /* Loop counters */
	int delcnt = 0;

	for(idx=0; idx < DESK_CACHESIZE; idx++)
		if(NULL != deskcache[idx])
		{
			gdk_pixmap_unref(deskcache[idx]);
			delcnt++;
		}
	
	#ifdef CACHE_DEBUG
	printf("cache_desktop_destroy(): free %d elements\n", delcnt);
	#endif

	cache_desktop_init();
}


/* Retrieve desktop data from the cache */
void* cache_desktop_get(uint32 offset)
{
	return deskcache[offset];
}

/* Store desktop data in the cache */
void cache_desktop_put(uint32 offset, void* data)
{
	assert((offset) <= sizeof(deskcache)/sizeof(void*));
	
	gdk_pixmap_ref(data);
	deskcache[offset] = data;
}

