/*
   rdesktop: A Remote Desktop Protocol client.
   User interface services - X-Windows
   Copyright (C) Matthew Chapman 1999-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/******************************************************************************
 * Name: xwin.c
 *
 * Purpose: To encapsulate calls to the GUI system.
 *
 * Notes:
 * 0. Modified under terms of the GPL by Luke Scharf (luke@vt.edu)
 *	1. Supports xlib and gtk+.  Hopefully svgalib will be added later.
 *	2. Hopefully, this file will be split into gui-gtk, gui-xlib, and
 *		gui-svgalib
 *
 *****************************************************************************/


/************
 * Includes 
 ***********/
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>
#include <time.h>
#include <assert.h>
#include "rdesktop.h"
#include "cache.h"


/************
 * Constants
 ***********/
#define FIXME 0
#define PATBLT_SOLID 0
#define PATBLT_PATTERN 3
#define UPDATE_SCREEN_MAX 32768

#define KEY_SHIFT_L 0x2A
#define KEY_SHIFT_R 0x36
#define KEY_CONTROL_L 0x1D
#define KEY_CONTROL_R 0x1D
#define KEY_ALT_L 0x38
#define KEY_ALT_R 0x38


static GdkFunction rop2_map[] = {
	/* 0 Clear */         GDK_CLEAR,        /* 0 */
	/* 1 nor */           GDK_SET,          /* DPon */
	/* 2 andInverted */   GDK_AND_INVERT,   /* DPna */
	/* 3 copyInverted */  GDK_COPY_INVERT,  /* Pn */
	/* 4 andReverse */    GDK_AND_REVERSE,  /* PDna */
	/* 5 invert */        GDK_INVERT,       /* Dn */
	/* 6 xor */           GDK_XOR,          /* DPx */
	/* 7 nand */          GDK_NAND,         /* DPan */
	/* 8 and */           GDK_AND,          /* DPa */
	/* 9 equiv */         GDK_EQUIV,        /* DPxn */
	/* A noop */          GDK_NOOP,         /* D */
	/* B orInverted */    GDK_OR_INVERT,          /* DPno */
	/* C copy */          GDK_COPY,         /* P */
	/* D orReverse */     GDK_OR_REVERSE,          /* PDno */
	/* E or */            GDK_OR,           /* DPo */
	/* F set */           GDK_SET           /* 1 */
};



/************
 * Globals
 ***********/
extern int width;
extern int height;
extern BOOL motion;

static GtkWidget* gtk_window;
static GtkWidget* gtk_canvas;
static GdkPixmap*	gdk_pixmap;
static HCOLOURMAP colormap;

static BOOL key_shift_l_down;
static BOOL key_shift_r_down;
static BOOL key_alt_l_down;
static BOOL key_alt_r_down;
static BOOL key_control_l_down;
static BOOL key_control_r_down;

static BOOL gtk_window_hasfocus;


/**************
 * Prototypes 
 *************/
gint gtk_configure_event(GtkWidget *widget, GdkEventConfigure *event, gpointer data);
gint gtk_expose_event(GtkWidget *widget, GdkEventExpose *event, gpointer data);
gint gtk_delete_event(GtkWidget *widget, GdkEvent  *event, gpointer data);

gboolean gtk_key_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data);
gboolean gtk_buttonpress_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
gboolean gtk_buttonrelease_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
gboolean gtk_motionnotify_event(GtkWidget *widget, GdkEventMotion *event, gpointer user_data);
gboolean gtk_gotfocus_event(GtkWidget *widget, GdkEventFocus   *event, gpointer user_data);
gboolean gtk_lostfocus_event(GtkWidget *widget, GdkEventFocus   *event, gpointer user_data);

static uint16 gtk_translate_mouse(unsigned long button);
uint8 gtk_translate_key(guint keyval);
void gtk_update_screen(int x, int y, int cx, int cy);


/******************
 * Name: ui_init
 *
 * Role: Exposed
 *****************/
void ui_init(int* argc_p, char*** argv_p)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_init\n");
	#endif

	/** GTK Initialization **/
	gtk_init(argc_p, argv_p);

	/** Cache Initialization **/
	cache_bmp_init();
	cache_font_init();
	cache_text_init();
	cache_desktop_init();
}


/******************
 * Name: ui_exit
 *
 * Role: Exposed
 *****************/
void ui_exit()
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_exit\n");
	#endif
	/** Cache Initialization **/
	cache_bmp_destroy();
	cache_font_destroy();
	cache_desktop_destroy();
	cache_text_destroy();

	/** GTK Shutdown */
	/*gtk_exit(0);*/
}



/***************************
 * Name: ui_create_window
 *
 * Role: Exposed
 **************************/
BOOL ui_create_window(char *title)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_create_window\n");
	#endif

	/* Misc */
	gtk_window_hasfocus = FALSE;

	/* Keyobaord */
	key_shift_l_down = FALSE;
	key_shift_r_down = FALSE;
	key_alt_l_down = FALSE;
	key_alt_r_down = FALSE;
	key_control_l_down = FALSE;
	key_control_r_down = FALSE;

	/* Double buffering and colors */
	gdk_pixmap = NULL;
	colormap = NULL;

	/* Create the top level Window */
	gtk_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_policy(GTK_WINDOW(gtk_window), FALSE, FALSE, TRUE);
	gtk_window_set_title(GTK_WINDOW(gtk_window), title);
	gtk_widget_set_events(GTK_WIDGET(gtk_window),
				 GDK_KEY_PRESS_MASK
				 | GDK_KEY_RELEASE_MASK);
	gtk_signal_connect(GTK_OBJECT(gtk_window),"key_press_event",
				(GtkSignalFunc)gtk_key_event, NULL);
	gtk_signal_connect(GTK_OBJECT(gtk_window),"key_release_event",
				(GtkSignalFunc)gtk_key_event, NULL);
	gtk_signal_connect(GTK_OBJECT(gtk_window),"delete_event",
				(GtkSignalFunc)gtk_delete_event, NULL);
	gtk_signal_connect(GTK_OBJECT(gtk_window),"focus_out_event",
				(GtkSignalFunc)gtk_lostfocus_event, NULL);
	gtk_signal_connect(GTK_OBJECT(gtk_window),"focus_in_event",
				(GtkSignalFunc)gtk_gotfocus_event, NULL);

	/* Create the canvas */
	gtk_canvas = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(gtk_canvas), width, height);
	gtk_widget_set_events(GTK_WIDGET(gtk_canvas),
             GDK_POINTER_MOTION_MASK
				 | GDK_BUTTON_PRESS_MASK
				 | GDK_BUTTON_RELEASE_MASK
				 | GDK_KEY_PRESS_MASK
				 | GDK_KEY_RELEASE_MASK);

	GTK_WIDGET_SET_FLAGS(gtk_canvas, GTK_CAN_FOCUS);
	gtk_signal_connect(GTK_OBJECT(gtk_canvas),"configure_event",
				(GtkSignalFunc)gtk_configure_event, NULL);
	gtk_signal_connect(GTK_OBJECT(gtk_canvas),"expose_event",
				(GtkSignalFunc)gtk_expose_event, NULL);
	gtk_signal_connect(GTK_OBJECT(gtk_canvas), "button_press_event",
				(GtkSignalFunc)gtk_buttonpress_event, NULL);
	gtk_signal_connect(GTK_OBJECT(gtk_canvas), "button_release_event",
				(GtkSignalFunc)gtk_buttonrelease_event, NULL);
	gtk_signal_connect(GTK_OBJECT(gtk_canvas),"motion_notify_event",
				(GtkSignalFunc)gtk_motionnotify_event, NULL);
	gtk_signal_connect(GTK_OBJECT(gtk_canvas),"key_press_event",
				(GtkSignalFunc)gtk_key_event, NULL);
	gtk_signal_connect(GTK_OBJECT(gtk_canvas),"key_release_event",
				(GtkSignalFunc)gtk_key_event, NULL);

	gtk_container_add(GTK_CONTAINER(gtk_window), gtk_canvas);

	/* Show everything */
	gtk_widget_show_all(gtk_window);

	/* return */
	return True;
}



/***************************
 * Name: ui_destroy_window
 *
 * Role: Exposed
 **************************/
void ui_destroy_window()
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_destroy_window\n");
	#endif

	/* Destroy */
	ui_destroy_colourmap(colormap);
	gtk_widget_destroy(gtk_canvas);
	gtk_widget_destroy(gtk_window);
	gdk_pixmap_unref(gdk_pixmap);
}


/***************************
 * Name: ui_process_events
 *
 * Role: Exposed
 **************************/
void ui_process_events()
{
	gtk_main_iteration_do(FALSE);	
	while (gtk_events_pending())
		gtk_main_iteration();
}


/***************************
 * Name: ui_warp_pointer
 *
 * Role: Exposed
 **************************/
void ui_move_pointer(int x, int y)
{
	#ifdef GUI_DEBUG
	printf("ui_move_pointer()\n");
	#endif

	if (FALSE != gtk_window_hasfocus)
	{
		XWarpPointer(GDK_WINDOW_XDISPLAY(gtk_canvas->window),None,
                GDK_WINDOW_XWINDOW(gtk_canvas->window),0,0,0,0,x,y);
	}
}



/***************************
 * Name: ui_create_bitmap
 *
 * Role: Exposed
 **************************/
HBITMAP ui_create_bitmap(int width, int height, uint8 *data)
{
	/* Variables */
	GdkPixmap* my_pixmap;
	int i,j;

	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_create_bitmap(%d, %d, data)\n", width, height);
	#endif

	/* Push image to X-Server */
	/* slow, but the colors are right. */
	gdk_gc_set_function(colormap->gc, GDK_COPY);
	my_pixmap = gdk_pixmap_new(gtk_canvas->window, width, height, -1);
	for(i=0; i<height; i++)
		for(j=0; j<width; j++)
		{
			gdk_gc_set_foreground(colormap->gc, colormap->colors[data[width*i+j]]);
			gdk_draw_point(my_pixmap, colormap->gc, j, i);
		}


	/* Cleanup */
	return (HBITMAP)my_pixmap;
}


/***************************
 * Name: ui_paint_bitmap
 *
 * Role: Exposed
 **************************/
void ui_paint_bitmap(int x, int y, int cx, int cy,
			int width, int height, uint8 *data)
{
	/* Variable Declarations */
	int i, j; /* Loop counters */

	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_paint_bitmap()\n");
	#endif

	/* Push image to X-Server */
	/* slow, but effective. */
	assert(cx <= width);
	assert(cy <= height);
	gdk_gc_set_function(colormap->gc, GDK_COPY);
	for(i=0; i<cy; i++)
		for(j=0; j<cx; j++)
		{
			gdk_gc_set_foreground(colormap->gc, colormap->colors[data[width*i+j]]);
			gdk_draw_point(gdk_pixmap, colormap->gc, x+j, y+i);
		}

	/* Update double-buffer */
	gtk_update_screen(x, y, cx, cy);
}


/***************************
 * Name: ui_destroy_bitmap
 *
 * Role: Exposed
 **************************/
void ui_destroy_bitmap(HBITMAP my_pixmap)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_destroy_bitmap\n");
	#endif

	/* Do it! */
	gdk_pixmap_unref(my_pixmap);
}


/*****************************************************************************
 * Name: ui_create_glyph
 *
 * Role: Exposed
 *
 * Notes: change this to use pixmaps at some point - GdkImages are too slow
 ****************************************************************************/
HGLYPH ui_create_glyph(int glyph_width, int glyph_height, uint8 *data)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_create_glyph()\n");
	#endif

	/* Create Bitmap */
	return gdk_bitmap_create_from_data(gtk_canvas->window,
									data, glyph_width, glyph_height);
}


/***************************
 * Name: ui_destroy_glyph
 *
 * Role: Exposed
 **************************/
void ui_destroy_glyph(HGLYPH glyph)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_destroy_glyph(): Begin\n");
	#endif

	/* Do it! */
	gdk_bitmap_unref(glyph);

	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_destroy_glyph(): End\n");
	#endif
}


/******************************
 * Name: ui_create_colourmap
 *
 * Role: Exposed
 *****************************/
HCOLOURMAP ui_create_colourmap(COLOURMAP *colours)
{
	/* Variable Declarations */
	HCOLOURMAP newmap = g_malloc(sizeof(HCOLOURMAP));
	int i;

	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_create_colourmap\n");
	#endif
	assert(NULL != colours);
	assert(NULL != newmap);

	/* setup */
	newmap->gc = gdk_gc_new(gtk_canvas->window);
	newmap->ncolors = colours->ncolours;
	newmap->colors = g_malloc(newmap->ncolors*sizeof(GdkColor*));
	assert(NULL != newmap->colors);
	newmap->colormap = gdk_colormap_get_system();

	/* Figure which colors we want */
	for (i = 0; i < newmap->ncolors; i++)
	{
		/* Pick the color */
		newmap->colors[i]        = g_malloc(sizeof(GdkColor));
		assert(NULL != newmap->colors[i]);
		newmap->colors[i]->pixel = i;
		newmap->colors[i]->red   = colours->colours[i].red*256;
		newmap->colors[i]->green = colours->colours[i].green*256;
		newmap->colors[i]->blue  = colours->colours[i].blue*256;

		gdk_colormap_alloc_color(newmap->colormap, newmap->colors[i], FALSE, TRUE);
	}

	/* Set misc attributes */
	gdk_gc_set_fill(newmap->gc, GDK_SOLID);
	gdk_gc_set_line_attributes(newmap->gc,
						 GUI_GTK_LINEWIDTH,
						 GDK_LINE_SOLID,
						 GDK_CAP_NOT_LAST,
						 GDK_JOIN_BEVEL);

	/* Return */
	return newmap;
}


/******************************
 * Name: ui_destroy_colourmap
 *
 * Role: Exposed
 *****************************/
void ui_destroy_colourmap(HCOLOURMAP map)
{
	/** Debug **/
	#ifdef GUI_DEBUG
	printf("ui_destroy_colourmap\n");
	#endif

	/* Assumes that we don't try to destroy the current colormap */
	gdk_colormap_free_colors(map->colormap, *(map->colors), map->ncolors);
	g_free(map->colors);
	/*free(map);*/

}


/******************************
 * Name: ui_set_colourmap
 *
 * Role: Exposed
 *****************************/
void ui_set_colourmap(HCOLOURMAP map)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_set_colourmap\n");
	#endif

	/* Install Colormap */
	colormap = map;
}


/******************************
 * Name: ui_set_clip
 *
 * Role: Exposed
 *****************************/
void ui_set_clip(int x, int y, int cx, int cy)
{
	/* Variable Declarations */
	GdkRectangle my_rect;
	
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_set_clip\n");
	#endif

	/* Do it! */
	my_rect.x = x;
	my_rect.y = y;
	my_rect.width = cx;
	my_rect.height = cy;
	gdk_gc_set_clip_rectangle(colormap->gc, &my_rect);
		
}


/******************************
 * Name: ui_reset_clip
 *
 * Role: Exposed
 *****************************/
void ui_reset_clip()
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_reset_clip\n");
	#endif

	/* Do it! */
	ui_set_clip(0,0, width,height);
}



/*********************
 * Name: ui_bell
 *
 * Role: Exposed
 ********************/
void ui_bell()
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_bell\n");
	#endif

	/* Do it! */
	gdk_beep();
}



/***********************
 * Name: ui_destblt
 *
 * Role: Exposed
 **********************/
void ui_destblt(uint8 opcode,
	/* dest */  int x, int y, int cx, int cy)
{
	/* Debug */
	#ifdef GUI_DEBUG	
	printf("ui_destblt\n");
	#endif

	/* Set the attributes */
	gdk_gc_set_function(colormap->gc, rop2_map[opcode]);
	gdk_gc_set_fill(colormap->gc, GDK_SOLID);
	gdk_gc_set_line_attributes(colormap->gc,
						 GUI_GTK_LINEWIDTH,
						 GDK_LINE_SOLID,
						 GDK_CAP_NOT_LAST,
						 GDK_JOIN_BEVEL);

	/* Draw the rectangle */
	gdk_draw_rectangle(gdk_pixmap, colormap->gc, TRUE,
                      x,y, cx,cy);

	
	/* Update the screen */
	gtk_update_screen(x, y, cx+2*GUI_GTK_LINEWIDTH, cy+2*GUI_GTK_LINEWIDTH);


}



/***********************
 * Name: ui_patblt
 *
 * Role: Exposed
 **********************/
void ui_patblt(uint8 opcode,
	/* dest */  int x, int y, int cx, int cy,
	/* brush */ BRUSH *brush, int bgcolour, int fgcolour)
{
	/* Variable Declarations */
	HGLYPH fill;

	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_patblt(): opcode = 0x%X\n", opcode);
	#endif

	/* Protections */
	assert(fgcolour < colormap->ncolors);
	assert(bgcolour < colormap->ncolors);

	/* Set the attributes */
	gdk_gc_set_foreground(colormap->gc, colormap->colors[fgcolour]);
	gdk_gc_set_background(colormap->gc, colormap->colors[bgcolour]);
	gdk_gc_set_function(colormap->gc, rop2_map[opcode]);


	/* Blit something Solid */	
	if (PATBLT_SOLID == brush->style)
	{
		/* Set Colors */
		gdk_gc_set_foreground(colormap->gc, colormap->colors[fgcolour]);
		gdk_gc_set_background(colormap->gc, colormap->colors[bgcolour]);

		/* Draw the rectangle */
		gdk_draw_rectangle(gdk_pixmap, colormap->gc, TRUE, x,y, cx,cy);
	}


	/* Blit in a pattern Pattern */
	else if (PATBLT_PATTERN == brush->style)
	{

		/* Draw the rectangle */
		gdk_draw_rectangle(gdk_pixmap, colormap->gc, TRUE, x,y, cx,cy);
		#ifdef GUI_DEBUG
		printf("ui_patblt():: Pattern\n");
		#endif

		fill = (HGLYPH)ui_create_glyph(8, 8, brush->pattern);

		gdk_gc_set_fill(colormap->gc, GDK_OPAQUE_STIPPLED);
		gdk_gc_set_stipple(colormap->gc, fill);

		gdk_draw_rectangle(gdk_pixmap, colormap->gc, TRUE, x,y, cx,cy);

		gdk_gc_set_fill(colormap->gc, GDK_SOLID);
		ui_destroy_glyph((HGLYPH)fill);
	}

	/* Just in case */
	else
	{
			NOTIMP("brush %d\n", brush->style);
	}


	/* Update the screen */
	gtk_update_screen(x,y, cx,cy);
}


/***********************
 * Name: ui_screenblt
 *
 * Role: Exposed
 **********************/
void ui_screenblt(uint8 opcode,
		/* dest */ int x, int y, int cx, int cy,
		/* src */  int srcx, int srcy)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_screenblt(): opcode = 0x%X\n", opcode);
	#endif

	/* copy visible screen -> visible screen */
	gdk_gc_set_function(colormap->gc, rop2_map[opcode]);
	gdk_draw_pixmap(gdk_pixmap, colormap->gc, gdk_pixmap,
                     srcx, srcy,
							x, y,
							cx, cy);

	/* Update double-buffer */
	gtk_update_screen(x, y, cx, cy);
}


/***********************
 * Name: ui_memblt
 *
 * Role: Exposed
 **********************/
void ui_memblt(uint8 opcode,
	/* dest */  int x, int y, int cx, int cy,
	/* src */   HBITMAP src, int srcx, int srcy)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_memblt(0x%X, %d, %d)\n", opcode, x, y);
	#endif

	/* Setup parameters */
	gdk_gc_set_function(colormap->gc, rop2_map[opcode]);

	/* Copy invisible screen -> visible screen */
	gdk_draw_pixmap(gdk_pixmap, colormap->gc, src,
						 srcx,srcy, x,y, cx,cy);

	/* Update double-buffer */
	gtk_update_screen(x, y, cx, cy);
}



/***********************
 * Name: ui_triblt
 *
 * Role: Exposed
 **********************/
void ui_triblt(uint8 opcode,
	/* dest */  int x, int y, int cx, int cy,
	/* src */   HBITMAP src, int srcx, int srcy,
	/* brush */ BRUSH *brush, int bgcolour, int fgcolour)
{
	/* This is potentially difficult to do in general. Until someone
	   comes up with a more efficient way of doing it I am using cases. */

	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_triblt()\n");
	#endif

	switch (opcode)
	{
		case 0x69: /* PDSxxn */
			ui_memblt(ROP2_XOR, x, y, cx, cy, src, srcx, srcy);
			ui_patblt(ROP2_NXOR, x, y, cx, cy,
					brush, bgcolour, fgcolour);
			break;

		case 0xb8: /* PSDPxax */
			ui_patblt(ROP2_XOR, x, y, cx, cy,
					brush, bgcolour, fgcolour);
			ui_memblt(ROP2_AND, x, y, cx, cy, src, srcx, srcy);
			ui_patblt(ROP2_XOR, x, y, cx, cy,
					brush, bgcolour, fgcolour);
			break;

		default:
			NOTIMP("triblt 0x%x\n", opcode);
			ui_memblt(ROP2_COPY, x, y, cx, cy, src, srcx, srcy);
	}
}


/*************************************
 * Name: ui_line
 *
 * Role: Exposed graphics primative
 ************************************/
void ui_line(uint8 opcode,
	/* dest */  int startx, int starty, int endx, int endy,
	/* pen */   PEN *pen)
{
	/* Variables */
	int left, right, top, bottom;

	/** Debug **/
	#ifdef GUI_DEBUG
	printf("ui_line(): opcode = 0x%X\n", opcode);
	#endif

	/* Set drawing attributes */
	assert(pen->colour < colormap->ncolors);
	gdk_gc_set_foreground(colormap->gc, colormap->colors[pen->colour]);
	gdk_gc_set_background(colormap->gc, colormap->colors[pen->colour]);
	gdk_gc_set_function(colormap->gc, rop2_map[opcode]);
	gdk_gc_set_fill(colormap->gc, GDK_SOLID);
	gdk_gc_set_line_attributes(colormap->gc,
						 GUI_GTK_LINEWIDTH,
						 GDK_LINE_SOLID,
						 GDK_CAP_NOT_LAST,
						 GDK_JOIN_BEVEL);
	
	/* Draw line */
	gdk_draw_line(	gdk_pixmap, colormap->gc,
                  startx, starty, endx, endy);

	/* Figure out where we drew the line */
	left   = (startx < endx)? startx : endx; left   = (left >= 0)?   left   : 0;
	right  = (startx > endx)? startx : endx; right  = (right >= 0)?  right  : 0;
	assert(left <= right);
	top    = (starty < endy)? starty : endy; top    = (top >= 0)?    top    : 0;
	bottom = (starty > endy)? starty : endy; bottom = (bottom >= 0)? bottom : 0;
	assert(top <= bottom);

	/* Update the screen */
	gtk_update_screen(left, top, right-left+2, bottom-top+2);
}



/*************************************
 * Name: ui_rect
 *
 * Role: Exposed graphics primative
 ************************************/
void ui_rect(
	/* dest */  int x, int y, int cx, int cy,
	/* brush */ int colour)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_rect(%d, %d, %d, %d, %d)\n", x, y, cx, cy, colour);
	#endif

	/* Set the attributes */
	assert(colour < colormap->ncolors);
	gdk_gc_set_foreground(colormap->gc, colormap->colors[colour]);
	gdk_gc_set_background(colormap->gc, colormap->colors[colour]);
	gdk_gc_set_function(colormap->gc, GDK_COPY);

	/* Draw the rectangle */
	gdk_draw_rectangle(gdk_pixmap, colormap->gc, TRUE,
                      x,y, cx,cy);

	
	/* Update the screen */
	gtk_update_screen(x, y, cx+2*GUI_GTK_LINEWIDTH, cy+2*GUI_GTK_LINEWIDTH);
}


/*************************************
 * Name: ui_draw_glyph
 *
 * Role: Exposed graphics primative
 ************************************/
void ui_draw_glyph(int mixmode,
	/* dest */ int x, int y, int cx, int cy,
	/* src */  HGLYPH glyph, int srcx, int srcy, int bgcolour, int fgcolour)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_draw_glyph(): fg = %d  bg = %d\n", fgcolour, bgcolour);
	#endif

	/* Idiot Protection */
	assert(fgcolour < colormap->ncolors);
	assert(bgcolour < colormap->ncolors);

	/* Setup */
	gdk_gc_set_function(colormap->gc, GDK_COPY);

	/* Opaque? Clear the area */
	if (MIX_OPAQUE == mixmode)
	{
		gdk_gc_set_foreground(colormap->gc, colormap->colors[bgcolour]);
		gdk_gc_set_background(colormap->gc, colormap->colors[fgcolour]);
		gdk_draw_rectangle(gdk_pixmap, colormap->gc, TRUE, x,y, cx,cy);
	}

	/* Draw transparent glyph */
	gdk_gc_set_foreground(colormap->gc, colormap->colors[fgcolour]);
	gdk_gc_set_background(colormap->gc, colormap->colors[bgcolour]);
	gdk_gc_set_stipple(colormap->gc, glyph);
	gdk_gc_set_fill(colormap->gc, GDK_STIPPLED);
	gdk_gc_set_ts_origin(colormap->gc, x, y);
	gdk_draw_rectangle(gdk_pixmap, colormap->gc, TRUE, x,y, cx,cy);
	gdk_gc_set_fill(colormap->gc, GDK_SOLID);

	/* Update the screen */
	gtk_update_screen(x, y, cx,cy);
}


/***********************
 * name: ui_draw_text
 *
 * Role: Exposed
 **********************/
void ui_draw_text(uint8 font, uint8 flags, int mixmode, int x, int y,
			int clipx, int clipy, int clipcx, int clipcy,
			int boxx, int boxy, int boxcx, int boxcy,
			int bgcolour, int fgcolour, uint8 *text, uint8 length)
{
	/* Debug */
	FONTGLYPH *glyph;
	int i;

	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_draw_text\n");
	#endif

	/* Clear background */
	if (boxcx > 1)
	{
		ui_rect(boxx, boxy, boxcx, boxcy, bgcolour);
	}
	else if (mixmode == MIX_OPAQUE)
	{
		ui_rect(clipx, clipy, clipcx, clipcy, bgcolour);
	}

	/* Paint text, character by character */
	for (i = 0; i < length; i++)
	{
		glyph = cache_font_get(font, text[i]);

		if (!(flags & TEXT2_IMPLICIT_X))
			x += text[++i];

		if (glyph != NULL)
		{
			ui_draw_glyph(mixmode, x + (short)glyph->offset,
					y + (short)glyph->baseline,
					glyph->width, glyph->height,
					glyph->pixmap, 0, 0,
					bgcolour, fgcolour);

			if (flags & TEXT2_IMPLICIT_X)
				x += glyph->width;
		}
	}
}



/*************************************
 * Name: ui_desktop_save
 *
 * Role: Exposed
 ************************************/
void ui_desktop_save(uint32 offset, int x, int y, int cx, int cy)
{
	/* Variable Declarations */
	GdkPixmap* my_pixmap;

	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_desktop_save\n");
	#endif

	/* Capture image */
	gdk_gc_set_function(colormap->gc, GDK_COPY);
	my_pixmap = gdk_pixmap_new(gtk_canvas->window, cx,cy, -1);
	gdk_draw_pixmap(my_pixmap, colormap->gc, gdk_pixmap,
						 x,y, 0,0, cx,cy);
						 
	cache_desktop_put(offset, (void*)my_pixmap);
}



/*************************************
 * Name: ui_desktop_restore
 *
 * Role: Exposed
 ************************************/
void ui_desktop_restore(uint32 offset, int x, int y, int cx, int cy)
{
	/* Variable Declarations */
	GdkPixmap* my_pixmap;

	/* Debug */
	#ifdef GUI_DEBUG
	printf("ui_desktop_restore\n");
	#endif

	/* Get pixmap from cache */
	my_pixmap = (GdkPixmap*)cache_desktop_get(offset);
	if (my_pixmap == NULL)
		return;
	
	/* Display it */
	gdk_gc_set_function(colormap->gc, GDK_COPY);
	gdk_draw_pixmap(gdk_pixmap, colormap->gc, my_pixmap,
						 0,0, x,y, cx, cy);
	gtk_update_screen(x, y, cx, cy);

	/* Cleanup */
	gdk_pixmap_unref(my_pixmap);
}





/*************************************
 * Name: gtk_configure_event
 *
 * Role: GTK Callback - gtk_canvas
 ************************************/
gint	gtk_configure_event(GtkWidget* widget, GdkEventConfigure* event,
                            gpointer data)
{
	/* debug */
	#ifdef GUI_DEBUG
	printf("gtk_configure_event()\n");
	#endif

	/* Delete old pixmap, if necessary */
	if (NULL != gdk_pixmap)
		gdk_pixmap_unref(gdk_pixmap);

	/* Create new pixmap */
	gdk_pixmap = gdk_pixmap_new(widget->window,
                          widget->allocation.width,
                          widget->allocation.height,
                          -1);

	/* Put Dummy image on pixmap */
	gdk_draw_rectangle(gdk_pixmap,
			gtk_canvas->style->white_gc,
			TRUE,
			1, 1,
			GTK_WIDGET(gtk_canvas)->allocation.width-2,
			GTK_WIDGET(gtk_canvas)->allocation.height-2);

	/* Return Success */
	return TRUE;
}



/********************************
 * Name: gtk_expose_event
 *
 * Role: GTK Callback - canvas
 *******************************/
gint	gtk_expose_event(GtkWidget* widget, GdkEventExpose* event,
                            gpointer data)
{
	/* debug */
	#ifdef GUI_DEBUG
	printf("gtk_expose_event()\n");
	#endif

	/* Paint Pixmap */
	gdk_draw_pixmap(widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			gdk_pixmap,
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height);

	/* Return Success */
	return TRUE;
}


/***************************
 * Name: gtk_delete_event
 *
 * Role: GTK Callback
 **************************/
gint	gtk_delete_event(GtkWidget* widget, GdkEvent* event,
                            gpointer data)
{
	/* debug */
	#ifdef GUI_DEBUG
	printf("gtk_delete_event()\n");
	#endif

	/* Shutdown */
	/* ui_exit(); */

	/* Return Success */
	return TRUE;
}


/***************************
 * Name: gtk_key_event
 *
 * Role: GTK Callback
 **************************/
gboolean gtk_key_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
	/* Variable Declarations */
	uint8 scancode = gtk_translate_key(event->keyval);
	uint32 mask = 0;

	/* debug */
	#ifdef GUI_DEBUG
	printf("gtk_key_event()\n");
	if (GDK_KEY_PRESS == event->type)
		printf("gtk_key_event() down keyval = 0x%X scancode = 0x%X ('%s')\n",
						 event->keyval,
						 scancode,
						 event->string);
		else
		printf("gtk_key_event() up keyval = 0x%X scancode = 0x%X ('%s')\n",
						 event->keyval,
						 scancode,
						 event->string);
	#endif

	/* Was there a key? */
	if (scancode == 0)
		return TRUE;

	/* classify tupe & record state */
	if (GDK_KEY_PRESS == event->type)
	{
		/* Record mask */
		mask = 0;

		/* Record key state */
		if(KEY_SHIFT_L == scancode) key_shift_l_down = TRUE;
		else if(KEY_SHIFT_R == scancode) key_shift_r_down = TRUE;
		else if(KEY_ALT_L == scancode) key_alt_l_down = TRUE;
		else if(KEY_ALT_R == scancode) key_alt_r_down = TRUE;
		else if(KEY_CONTROL_L == scancode) key_control_l_down = TRUE;
		else if(KEY_CONTROL_R == scancode) key_control_r_down = TRUE;
	}
	else
	{
		/* Record mask */
		mask = KBD_FLAG_DOWN | KBD_FLAG_UP;

		/* Record key state */
		if(KEY_SHIFT_L == scancode) key_shift_l_down = FALSE;
		else if(KEY_SHIFT_R == scancode) key_shift_r_down = FALSE;
		else if(KEY_ALT_L == scancode) key_alt_l_down = FALSE;
		else if(KEY_ALT_R == scancode) key_alt_r_down = FALSE;
		else if(KEY_CONTROL_L == scancode) key_control_l_down = FALSE;
		else if(KEY_CONTROL_R == scancode) key_control_r_down = FALSE;
	}

	/* Send it along! */
	rdp_send_input(event->time, RDP_INPUT_SCANCODE, mask, scancode, 0);

	/* Housekeeping */
	return TRUE;
}


/********************************
 * Name: gtk_buttonpress_event
 *
 * Role: GTK Callback
 *******************************/
gboolean gtk_buttonpress_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
	/* Variable Declarations */
	uint16 button = gtk_translate_mouse(event->button);

	/* debug */
	#ifdef GUI_DEBUG
	printf("gtk_buttonpress_event()\n");
	#endif

	/* If there really wasn't a buttonpress */
	if (button == 0)
		return TRUE;

	/* Send the buttonpress along! */
	rdp_send_input(event->time, RDP_INPUT_MOUSE,
						button | MOUSE_FLAG_DOWN,
						event->x,
						event->y);

	return TRUE;
}


/**********************************
 * Name: gtk_buttonrelease_event
 *
 * Role: GTK Callback
 *********************************/
gboolean gtk_buttonrelease_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
	/* Variable Declarations */
	uint16 button = gtk_translate_mouse(event->button);

	/* debug */
	#ifdef GUI_DEBUG
	printf("gtk_buttonrelease_event()\n");
	#endif

	/* If there really wasn't a buttonpress */
	if (button == 0)
		return TRUE;

	/* Send the buttonpress along! */
	rdp_send_input(event->time, RDP_INPUT_MOUSE,
						button,
						event->x,
						event->y);

	/* Housekeeping */
	return TRUE;
}


/*********************************
 * Name: gtk_motionnotify_event
 *
 * Role: GTK Callback
 ********************************/
gboolean gtk_motionnotify_event(GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
	/* Send it along! */
	rdp_send_input(event->time, RDP_INPUT_MOUSE,
						MOUSE_FLAG_MOVE,
						event->x,
						event->y);

	/* Housekeeping */
	return TRUE;
}


/*******************************************************
 * Name: gtk_gotfocus_event
 *
 * Role: Internal
 *
 * Purpose: Ask gtk to update a region of the screen.
 ******************************************************/
gboolean gtk_gotfocus_event(GtkWidget *widget, GdkEventFocus   *event, gpointer user_data)
{
	/* Debug */
	#ifdef GUI_DEBUG
	printf("gtk_gotfocus_event()\n");
	#endif

	/* We're allowed to warp the pointer now */
	gtk_window_hasfocus = TRUE;

	/* Return */
	return TRUE;

}



/*******************************************************
 * Name: gtk_lostfocus_event
 *
 * Role: Internal
 *
 * Purpose: Ask gtk to update a region of the screen.
 ******************************************************/
gboolean gtk_lostfocus_event(GtkWidget *widget, GdkEventFocus   *event, gpointer user_data)
{
	/* Variable Declarations */
	GdkEventKey key_event;

	/* Debug */
	#ifdef GUI_DEBUG
	printf("gtk_lost_focus_event()\n");
	printf("gtk_lost_focus_event(): key_shift_l_down = %d\n", key_shift_l_down);
	printf("gtk_lost_focus_event(): key_shift_r_down = %d\n", key_shift_r_down);
	printf("gtk_lost_focus_event(): key_alt_l_down = %d\n", key_alt_l_down);
	printf("gtk_lost_focus_event(): key_alt_r_down = %d\n", key_alt_r_down);
	printf("gtk_lost_focus_event(): key_control_l_down = %d\n", key_control_l_down);
	printf("gtk_lost_focus_event(): key_control_r_down = %d\n", key_control_r_down);
	#endif

	/* Make sure we don't warp the pointer anywhere */
	gtk_window_hasfocus = FALSE;

	/* Modifier-keys */
	if (TRUE == key_shift_l_down)
	{
		key_event.type       = GDK_KEY_RELEASE;
		key_event.window     = gtk_window->window;
		key_event.send_event = TRUE;
		key_event.time       = -1;
		key_event.state		= 0;
		key_event.keyval     = GDK_Shift_L;
		key_event.length     = 0;
		key_event.string     = NULL;
		gtk_widget_event(GTK_WIDGET(gtk_window), (GdkEvent*)&key_event);
	}

	if (TRUE == key_shift_r_down)
	{
		key_event.type       = GDK_KEY_RELEASE;
		key_event.window     = gtk_window->window;
		key_event.send_event = TRUE;
		key_event.time       = -1;
		key_event.state		= 0;
		key_event.keyval     = GDK_Shift_R;
		key_event.length     = 0;
		key_event.string     = NULL;
		gtk_widget_event(GTK_WIDGET(gtk_window), (GdkEvent*)&key_event);
	}

	if (TRUE == key_alt_l_down)
	{
		key_event.type       = GDK_KEY_RELEASE;
		key_event.window     = gtk_window->window;
		key_event.send_event = TRUE;
		key_event.time       = -1;
		key_event.state		= 0;
		key_event.keyval     = GDK_Alt_L;
		key_event.length     = 0;
		key_event.string     = NULL;
		gtk_widget_event(GTK_WIDGET(gtk_window), (GdkEvent*)&key_event);
	}

	if (TRUE == key_alt_r_down)
	{
		key_event.type       = GDK_KEY_RELEASE;
		key_event.window     = gtk_window->window;
		key_event.send_event = TRUE;
		key_event.time       = -1;
		key_event.state		= 0;
		key_event.keyval     = GDK_Alt_R;
		key_event.length     = 0;
		key_event.string     = NULL;
		gtk_widget_event(GTK_WIDGET(gtk_window), (GdkEvent*)&key_event);
	}

	if (TRUE == key_control_l_down)
	{
		key_event.type       = GDK_KEY_RELEASE;
		key_event.window     = gtk_window->window;
		key_event.send_event = TRUE;
		key_event.time       = -1;
		key_event.state		= 0;
		key_event.keyval     = GDK_Control_L;
		key_event.length     = 0;
		key_event.string     = NULL;
		gtk_widget_event(GTK_WIDGET(gtk_window), (GdkEvent*)&key_event);
	}

	if (TRUE == key_control_r_down)
	{
		key_event.type       = GDK_KEY_RELEASE;
		key_event.window     = gtk_window->window;
		key_event.send_event = TRUE;
		key_event.time       = -1;
		key_event.state		= 0;
		key_event.keyval     = GDK_Control_R;
		key_event.length     = 0;
		key_event.string     = NULL;
		gtk_widget_event(GTK_WIDGET(gtk_window), (GdkEvent*)&key_event);
	}

	#ifdef GUI_DEBUG
	printf("-----------------\n");
	printf("gtk_lost_focus_event(): key_shift_l_down = %d\n", key_shift_l_down);
	printf("gtk_lost_focus_event(): key_shift_r_down = %d\n", key_shift_r_down);
	printf("gtk_lost_focus_event(): key_alt_l_down = %d\n", key_alt_l_down);
	printf("gtk_lost_focus_event(): key_alt_r_down = %d\n", key_alt_r_down);
	printf("gtk_lost_focus_event(): key_control_l_down = %d\n", key_control_l_down);
	printf("gtk_lost_focus_event(): key_control_r_down = %d\n", key_control_r_down);
	#endif

	/* Cleanup */
	return TRUE;
}


/*******************************************************
 * Name: gtk_update_screen
 *
 * Role: Internal
 *
 * Purpose: Ask gtk to update a region of the screen.
 ******************************************************/
void gtk_update_screen(int x, int y, int cx, int cy)
{
	/* Variable Declarations */
	GdkEventExpose event;

	/* Debug */
	#ifdef GUI_DEBUG
	printf("gtk_update_screen(%d,%d,%d,%d)\n",x,y,cx,cy);
	#endif

	/* Range Check */
	if (cx >= UPDATE_SCREEN_MAX)
	{
		x  = 0;
		cx = cx - UPDATE_SCREEN_MAX;
		assert(cx < UPDATE_SCREEN_MAX);
	}
	if (cy >= UPDATE_SCREEN_MAX)
	{
		y  = 0;
		cy = cy - UPDATE_SCREEN_MAX;
		assert(cy < UPDATE_SCREEN_MAX);
	}

	/* Compose */
	event.type        = GDK_EXPOSE;
	event.window      = gtk_canvas->window;
	event.send_event  = TRUE;
	event.area.x      = x;
	event.area.y      = y;
	event.area.width  = cx;
	event.area.height = cy;
	event.count       = 0;

	/* send event */
	gtk_widget_event(GTK_WIDGET(gtk_canvas), (GdkEvent*)&event);
}


/**********************************
 * Name: gtk_translate_mouse
 *
 * Role: GTK Callback helper
 *********************************/
static uint16 gtk_translate_mouse(unsigned long button)
{
	switch (button)
	{
		case Button1: /* left */
			return MOUSE_FLAG_BUTTON1;
		#ifdef GUI_3BUTTON
		case Button2: /* middle */
			return MOUSE_FLAG_BUTTON3;
		#endif
		case Button3: /* right */
			return MOUSE_FLAG_BUTTON2;
		default:
			return 0; /* Default to no button*/
	}

	return 0;
}


/***********************************************************************
 * Name: gtk_translate_key
 *
 * Role: GTK Callback helper
 *
 * Purpose: Map GTK+ keyvals to the PC scancodes that RDP expects.
 * 	Just one big-ass case statement.
 *********************************************************************/
uint8 gtk_translate_key(guint keyval)
{
	switch(keyval)
	{
		case GDK_Escape: return 0x01;
		case GDK_BackSpace: return 0x0E;
		case GDK_Return: return 0x1C;
		case GDK_Tab: return 0x0F;
		case GDK_ISO_Left_Tab: return 0x0F;
		case GDK_Home: return 0x47;
		case GDK_Left: return 0x4B;
		case GDK_Up: return 0x48;
		case GDK_Right: return 0x4D;
		case GDK_Down: return 0x50;
		case GDK_Page_Up: return 0x49;
		case GDK_Page_Down: return 0x51;
		case GDK_End: return 0x4F;
		case GDK_Insert: return 0x52;
		case GDK_Delete: return 0x53;
		case GDK_Num_Lock: return 0x45;
		case GDK_KP_Enter: return 0x1C;
		case GDK_KP_Home: return 0x47;
		case GDK_KP_Left: return 0x4B;
		case GDK_KP_Up: return 0x48;
		case GDK_KP_Right: return 0x4D;
		case GDK_KP_Down: return 0x50;
		case GDK_KP_Page_Up: return 0x49;
		case GDK_KP_Page_Down: return 0x51;
		case GDK_KP_End: return 0x4F;
		case GDK_KP_Insert: return 0x52;
		case GDK_KP_Delete: return 0x53;
		case GDK_KP_Multiply: return 0x37;
		case GDK_KP_Add: return 0x4E;
		case GDK_KP_Subtract: return 0x4A;
		case GDK_KP_Decimal: return 0x53;
		case GDK_KP_Divide: return 0x35;
		case GDK_KP_0: return 0x52;
		case GDK_KP_1: return 0x4F;
		case GDK_KP_2: return 0x50;
		case GDK_KP_3: return 0x51;
		case GDK_KP_4: return 0x4B;
		case GDK_KP_5: return 0x4C;
		case GDK_KP_6: return 0x4D;
		case GDK_KP_7: return 0x47;
		case GDK_KP_8: return 0x48;
		case GDK_KP_9: return 0x49;
		case GDK_F1: return 0x3B;
		case GDK_F2: return 0x3C;
		case GDK_F3: return 0x3D;
		case GDK_F4: return 0x3E;
		case GDK_F5: return 0x3F;
		case GDK_F6: return 0x40;
		case GDK_F7: return 0x41;
		case GDK_F8: return 0x42;
		case GDK_F9: return 0x43;
		case GDK_F10: return 0x44;
		case GDK_F11: return 0x57;
		case GDK_F12: return 0x58;
		case GDK_Shift_L: return KEY_SHIFT_L;
		case GDK_Shift_R: return KEY_SHIFT_R;
		case GDK_Control_L: return KEY_CONTROL_L;
		case GDK_Control_R: return KEY_CONTROL_R;
		case GDK_Caps_Lock: return 0x3A;
		case GDK_Alt_L: return KEY_ALT_L;
		case GDK_Alt_R: return KEY_ALT_R;
		case GDK_space: return 0x39;
		case GDK_exclam: return 0x02;
		case GDK_quotedbl: return 0x28;
		case GDK_numbersign: return 0x04;
		case GDK_dollar: return 0x05;
		case GDK_percent: return 0x06;
		case GDK_ampersand: return 0x08;
		case GDK_apostrophe: return 0x28;
		case GDK_parenleft: return 0x0A;
		case GDK_parenright: return 0x0B;
		case GDK_asterisk: return 0x09;
		case GDK_plus: return 0x0D;
		case GDK_comma: return 0x33;
		case GDK_minus: return 0x0C;
		case GDK_period: return 0x34;
		case GDK_slash: return 0x35;
		case GDK_1: return 0x02;
		case GDK_2: return 0x03;
		case GDK_3: return 0x04;
		case GDK_4: return 0x05;
		case GDK_5: return 0x06;
		case GDK_6: return 0x07;
		case GDK_7: return 0x08;
		case GDK_8: return 0x09;
		case GDK_9: return 0x0A;
		case GDK_0: return 0x0B;
		case GDK_colon: return 0x27;
		case GDK_semicolon: return 0x27;
		case GDK_less: return 0x33;
		case GDK_equal: return 0x0D;
		case GDK_greater: return 0x34;
		case GDK_question: return 0x35;
		case GDK_at: return 0x03;
		case GDK_A: return 0x1E;
		case GDK_B: return 0x30;
		case GDK_C: return 0x2E;
		case GDK_D: return 0x20;
		case GDK_E: return 0x12;
		case GDK_F: return 0x21;
		case GDK_G: return 0x22;
		case GDK_H: return 0x23;
		case GDK_I: return 0x17;
		case GDK_J: return 0x24;
		case GDK_K: return 0x25;
		case GDK_L: return 0x26;
		case GDK_M: return 0x32;
		case GDK_N: return 0x31;
		case GDK_O: return 0x18;
		case GDK_P: return 0x19;
		case GDK_Q: return 0x10;
		case GDK_R: return 0x13;
		case GDK_S: return 0x1F;
		case GDK_T: return 0x14;
		case GDK_U: return 0x16;
		case GDK_V: return 0x2F;
		case GDK_W: return 0x11;
		case GDK_X: return 0x2D;
		case GDK_Y: return 0x15;
		case GDK_Z: return 0x2C;
		case GDK_a: return 0x1E;
		case GDK_b: return 0x30;
		case GDK_c: return 0x2E;
		case GDK_d: return 0x20;
		case GDK_e: return 0x12;
		case GDK_f: return 0x21;
		case GDK_g: return 0x22;
		case GDK_h: return 0x23;
		case GDK_i: return 0x17;
		case GDK_j: return 0x24;
		case GDK_k: return 0x25;
		case GDK_l: return 0x26;
		case GDK_m: return 0x32;
		case GDK_n: return 0x31;
		case GDK_o: return 0x18;
		case GDK_p: return 0x19;
		case GDK_q: return 0x10;
		case GDK_r: return 0x13;
		case GDK_s: return 0x1F;
		case GDK_t: return 0x14;
		case GDK_u: return 0x16;
		case GDK_v: return 0x2F;
		case GDK_w: return 0x11;
		case GDK_x: return 0x2D;
		case GDK_y: return 0x15;
		case GDK_z: return 0x2C;
		case GDK_bracketleft: return 0x1A;
		case GDK_backslash: return 0x2B;
		case GDK_bracketright: return 0x1B;
		case GDK_underscore: return 0x0C;
		case GDK_bar: return 0x2B;
		case GDK_braceleft: return 0x1A;
		case GDK_braceright: return 0x1B;
		case GDK_asciitilde: return 0x29;
		case GDK_hyphen: return 0x0C;
		default:
			#ifdef GUI_DEBUG
			printf("Unknown keyval - 0x%X\n", keyval);
			#endif
			return 0;
	}
}


